import os
from datetime import timedelta
from enum import Enum
from pathlib import Path
from typing import Generic, Optional, TypeVar

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import tomli
import tomli_w
from pydantic import BaseModel, model_validator

from flood_adapt.misc.path_builder import TopLevelDir, db_path
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.csv import read_csv
from flood_adapt.objects.forcing.time_frame import REFERENCE_TIME, TimeFrame

TValueUnitPair = TypeVar("TValueUnitPair", bound=us.ValueUnitPair)


class ShapeType(str, Enum):
    gaussian = "gaussian"
    block = "block"
    triangle = "triangle"
    scs = "scs"


class Scstype(str, Enum):
    type1 = "type_1"
    type1a = "type_1a"
    type2 = "type_2"
    type3 = "type_3"


class SyntheticTimeseries(BaseModel):
    # Required
    shape_type: ShapeType
    duration: us.UnitfulTime
    peak_time: us.UnitfulTime

    # Either one of these must be set
    peak_value: Optional[us.ValueUnitPairs] = None
    cumulative: Optional[us.ValueUnitPairs] = None

    # Optional
    fill_value: float = 0.0

    @model_validator(mode="after")
    def positive_duration(self):
        if self.duration.value < 0:
            raise ValueError(
                f"Timeseries shape duration must be positive, got {self.duration}"
            )
        return self

    @model_validator(mode="after")
    def either_value_or_cumulative(self):
        if (self.peak_value is None and self.cumulative is None) or (
            self.peak_value is not None and self.cumulative is not None
        ):
            raise ValueError(
                "Either `peak_value` or `cumulative` must be specified for Synthetic Timeseries."
            )
        return self

    @property
    def start_time(self) -> us.UnitfulTime:
        return self.peak_time - self.duration / 2

    @property
    def end_time(self) -> us.UnitfulTime:
        return self.peak_time + self.duration / 2

    def calculate_data(
        self, time_step: timedelta = TimeFrame().time_step
    ) -> np.ndarray:
        """Interpolate timeseries data as a numpy array with the provided time step and time as index and intensity as column."""
        # @abstractmethod doesnt work nicely with pydantic BaseModel, so we use this instead
        raise NotImplementedError(
            "This method should be implemented in subclasses of SyntheticTimeseries."
        )

    def to_dataframe(
        self,
        time_frame: TimeFrame,
    ) -> pd.DataFrame:
        """
        Interpolate the timeseries data using the time_step provided.

        Parameters
        ----------
        start_time : datetime | str
            Start time of the timeseries.
        end_time : datetime | str
            End time of the timeseries.
        time_step : us.UnitfulTime, optional
            Time step of the timeseries, by default TimeFrame().time_step.

        """
        return self._to_dataframe(
            time_frame=time_frame,
            ts_start_time=self.start_time,
            ts_end_time=self.end_time,
            fill_value=self.fill_value,
        )

    def _to_dataframe(
        self,
        time_frame: TimeFrame,
        ts_start_time: us.UnitfulTime,
        ts_end_time: us.UnitfulTime,
        fill_value: float = 0.0,
    ) -> pd.DataFrame:
        """
        Convert timeseries data to a pandas DataFrame that has time as the index and intensity as the column.

        The dataframe time range is from start_time to end_time with the provided time_step.
        The timeseries data is added to this range by first
            - Interpolating the data to the time_step
            - Filling the missing values with 0.

        Args:
            time_frame (TimeFrame):
                The time frame for the data.
            ts_start_time (us.UnitfulTime):
                The start time of the timeseries data relative to the time_frame start time.
            ts_end_time (us.UnitfulTime):
                The end time of the timeseries data relative to the time_frame start time.
            fill_value (float, optional):
                The fill value for missing data. Defaults to 0.0.

        Returns
        -------
            pd.DataFrame: A pandas DataFrame with time as the index and values as the columns.
            The data is interpolated to the time_step and values that fall outside of the timeseries data are filled with 0.
        """
        full_df_time_range = pd.date_range(
            start=time_frame.start_time,
            end=time_frame.end_time,
            freq=time_frame.time_step,
            name="time",
        )

        data = self.calculate_data(time_step=time_frame.time_step) + fill_value

        n_cols = data.shape[1] if len(data.shape) > 1 else 1
        ts_time_range = pd.date_range(
            start=(time_frame.start_time + ts_start_time.to_timedelta()),
            end=(time_frame.start_time + ts_end_time.to_timedelta()),
            freq=time_frame.time_step,
        )

        # If the data contains more than the requested time range (from reading a csv file)
        # Slice the data to match the expected time range
        if len(data) > len(ts_time_range):
            data = data[: len(ts_time_range)]

        df = pd.DataFrame(
            data, columns=[f"data_{i}" for i in range(n_cols)], index=ts_time_range
        )

        full_df = df.reindex(
            index=full_df_time_range,
            method="nearest",
            limit=1,
            fill_value=fill_value,
        )
        full_df = full_df.set_index(full_df_time_range)
        full_df.index = pd.to_datetime(full_df.index)
        full_df.index.name = "time"
        return full_df

    @classmethod
    def load_file(cls, file_path: Path | str | os.PathLike) -> "SyntheticTimeseries":
        """Load object from file."""
        with open(file_path, mode="rb") as fp:
            toml = tomli.load(fp)
        return cls(**toml)

    def save(self, filepath: Path):
        """
        Save Synthetic Timeseries toml.

        Parameters
        ----------
        file : Path
            path to the location where file will be saved
        """
        with open(filepath, "wb") as f:
            tomli_w.dump(self.model_dump(exclude_none=True), f)

    @staticmethod
    def plot(
        df,
        xmin: pd.Timestamp,
        xmax: pd.Timestamp,
        timeseries_variable: us.ValueUnitPair,
    ) -> go.Figure:
        fig = px.line(data_frame=df)
        fig.update_layout(
            autosize=False,
            height=100 * 2,
            width=280 * 2,
            margin={"r": 0, "l": 0, "b": 0, "t": 0},
            font={"size": 10, "color": "black", "family": "Arial"},
            title_font={"size": 10, "color": "black", "family": "Arial"},
            legend=None,
            yaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
            xaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
            xaxis_title={"text": "Time"},
            yaxis_title={"text": f"{timeseries_variable.units}"},
            showlegend=False,
            xaxis={"range": [xmin, xmax]},
        )
        return fig

    def __eq__(self, other) -> bool:
        if not isinstance(other, SyntheticTimeseries):
            raise NotImplementedError(f"Cannot compare Timeseries to {type(other)}")

        # If the following equation is element-wise True, then allclose returns True.:
        # absolute(a - b) <= (atol + rtol * absolute(b))
        return np.allclose(
            self.calculate_data(),
            other.calculate_data(),
            rtol=1e-2,
        )


class ScsTimeseries(SyntheticTimeseries):
    shape_type: ShapeType = ShapeType.scs

    scs_file_name: str
    scs_type: Scstype

    def calculate_data(
        self, time_step: timedelta = TimeFrame().time_step
    ) -> np.ndarray:
        _duration = self.duration.convert(us.UnitTypesTime.seconds)
        _start_time = self.start_time.convert(us.UnitTypesTime.seconds)

        scs_df = pd.read_csv(
            db_path(top_level_dir=TopLevelDir.static) / "scs" / self.scs_file_name,
            index_col=0,
        )[self.scs_type]

        tt = pd.date_range(
            start=(REFERENCE_TIME + self.start_time.to_timedelta()),
            end=(REFERENCE_TIME + self.end_time.to_timedelta()),
            freq=time_step,
        )
        tt = (tt - REFERENCE_TIME).total_seconds()

        tt_rain = _start_time + scs_df.index.to_numpy() * _duration
        rain_series = scs_df.to_numpy()
        rain_instantaneous = np.diff(rain_series) / np.diff(
            tt_rain / 3600
        )  # divide by time in hours to get mm/hour

        # interpolate instanetaneous rain intensity timeseries to tt
        rain_interp = np.interp(
            tt,
            tt_rain,
            np.concatenate(([0], rain_instantaneous)),
            left=0,
            right=0,
        )
        rainfall = (
            rain_interp * self.cumulative.value / np.trapz(rain_interp, tt / 3600)
        )

        return rainfall

    @model_validator(mode="after")
    def validate_attrs(self):
        if self.cumulative is None:
            raise ValueError(
                f"SCS timeseries must have `cumulative` specified. {self.cumulative}"
            )
        return self


class GaussianTimeseries(SyntheticTimeseries):
    shape_type: ShapeType = ShapeType.gaussian

    def calculate_data(
        self, time_step: timedelta = TimeFrame().time_step
    ) -> np.ndarray:
        _start = self.start_time.convert(us.UnitTypesTime.hours)
        _end = self.end_time.convert(us.UnitTypesTime.hours)

        tt = pd.date_range(
            start=(REFERENCE_TIME + self.start_time.to_timedelta()),
            end=(REFERENCE_TIME + self.end_time.to_timedelta()),
            freq=time_step,
        )
        tt_hours = (tt - REFERENCE_TIME).total_seconds() / 3600

        mean = (_start + _end) / 2
        sigma = (_end - _start) / 6
        gaussian_curve = np.exp(-0.5 * ((tt_hours - mean) / sigma) ** 2)

        if self.cumulative:
            # Normalize to ensure the integral sums to 1 over the time steps
            integral_approx = np.trapz(gaussian_curve, tt_hours)
            normalized_gaussian = gaussian_curve / integral_approx
            ts = self.cumulative.value * normalized_gaussian.to_numpy()
        elif self.peak_value:
            ts = self.peak_value.value * gaussian_curve
        else:
            raise ValueError("Either peak_value or cumulative must be specified.")

        return ts

    @model_validator(mode="after")
    def validate_attrs(self):
        # either peak_value or cumulative must be set, which is already checked in the parent class: `either_value_or_cumulative`
        return self


class BlockTimeseries(SyntheticTimeseries):
    shape_type: ShapeType = ShapeType.block

    def calculate_data(
        self, time_step: timedelta = TimeFrame().time_step
    ) -> np.ndarray:
        tt = pd.date_range(
            start=(REFERENCE_TIME + self.start_time.to_timedelta()),
            end=(REFERENCE_TIME + self.end_time.to_timedelta()),
            freq=time_step,
        )
        if self.peak_value:
            height_value = self.peak_value.value
        elif self.cumulative:
            area = self.cumulative.value
            base = self.duration.convert(
                us.UnitTypesTime.hours
            )  # always expect duration in hours
            height_value = area / base

        ts = np.zeros((len(tt),)) + height_value
        return ts

    @model_validator(mode="after")
    def validate_attrs(self):
        # either peak_value or cumulative must be set, which is already checked in the parent class: `either_value_or_cumulative`
        return self


class TriangleTimeseries(SyntheticTimeseries):
    shape_type: ShapeType = ShapeType.triangle

    def calculate_data(
        self, time_step: timedelta = TimeFrame().time_step
    ) -> np.ndarray:
        tt = pd.date_range(
            start=(REFERENCE_TIME + self.start_time.to_timedelta()),
            end=(REFERENCE_TIME + self.end_time.to_timedelta()),
            freq=time_step,
        )
        tt_seconds = (tt - REFERENCE_TIME).total_seconds()
        peak_time = self.peak_time.to_timedelta().total_seconds()
        start_time = self.start_time.to_timedelta().total_seconds()

        if self.peak_value:
            height_value = self.peak_value.value
        elif self.cumulative:
            area = self.cumulative.value
            base = self.duration.convert(
                us.UnitTypesTime.hours
            )  # always expect duration in hours
            height_value = (2 * area) / base

        ascending_slope = (
            height_value
            / (self.peak_time - self.start_time).to_timedelta().total_seconds()
        )
        descending_slope = (
            -height_value
            / (self.end_time - self.peak_time).to_timedelta().total_seconds()
        )

        ts = np.piecewise(
            tt_seconds,
            [tt_seconds < peak_time, tt_seconds >= peak_time],
            [
                lambda x: np.maximum(ascending_slope * (x - start_time), 0),
                lambda x: np.maximum(
                    descending_slope * (x - peak_time) + height_value, 0
                ),
                0,
            ],
        )
        return ts

    @model_validator(mode="after")
    def validate_attrs(self):
        # either peak_value or cumulative must be set, which is already checked in the parent class: `either_value_or_cumulative`
        return self


class CSVTimeseries(BaseModel, Generic[TValueUnitPair]):
    path: Path
    units: TValueUnitPair

    @model_validator(mode="after")
    def validate_csv(self):
        if not self.path.exists():
            raise ValueError(f"Path {self.path} does not exist.")
        if not self.path.is_file():
            raise ValueError(f"Path {self.path} is not a file.")
        if not self.path.suffix == ".csv":
            raise ValueError(f"Path {self.path} is not a csv file.")

        # Try loading the csv file, read_csv will raise an error if it cannot read the file
        read_csv(self.path)
        return self

    @staticmethod
    def load_file(path: str | Path, units: us.ValueUnitPair):
        return CSVTimeseries[type(units)](path=Path(path), units=units)

    def to_dataframe(
        self,
        time_frame: TimeFrame,
        fill_value: float = 0,
    ) -> pd.DataFrame:
        """
        Interpolate the timeseries data using the time_step provided.

        Parameters
        ----------
        time_frame : TimeFrame
            Time frame for the data.
        fill_value : float, optional
            Value to fill missing data with, by default 0.

        Returns
        -------
        pd.DataFrame
            Interpolated timeseries with datetime index.
        """
        file_data = read_csv(self.path)

        # Ensure requested time range is within available data
        start_time = max(time_frame.start_time, file_data.index.min())
        end_time = min(time_frame.end_time, file_data.index.max())

        df = file_data.loc[start_time:end_time]

        # Generate the complete time range
        time_range = pd.date_range(
            start=time_frame.start_time,
            end=time_frame.end_time,
            freq=time_frame.time_step,
        )

        # Reindex and fill missing values with specified fill_value
        interpolated_df = (
            df.reindex(time_range, method="nearest", limit=1)
            .interpolate(method="linear")
            .fillna(fill_value)
        )
        interpolated_df.index.name = "time"
        return interpolated_df

    def calculate_data(
        self,
        time_step: timedelta = TimeFrame().time_step,
    ) -> np.ndarray:
        return read_csv(self.path).to_numpy()

    def read_time_frame(self) -> TimeFrame:
        """
        Read the time frame from the file.

        Returns
        -------
        TimeFrame
            Time frame of the data in the file.
        """
        file_data = read_csv(self.path)
        return TimeFrame(
            start_time=file_data.index.min(),
            end_time=file_data.index.max(),
        )


class TimeseriesFactory:
    @staticmethod
    def from_args(
        shape_type: ShapeType,
        duration: us.UnitfulTime,
        peak_time: us.UnitfulTime,
        peak_value: Optional[us.ValueUnitPairs] = None,
        cumulative: Optional[us.ValueUnitPairs] = None,
        fill_value: float = 0.0,
        scs_file_name: Optional[str] = None,
        scs_type: Optional[Scstype] = None,
    ) -> SyntheticTimeseries:
        """Create a timeseries object based on the shape type."""
        match shape_type:
            case ShapeType.gaussian:
                return GaussianTimeseries(
                    duration=duration,
                    peak_time=peak_time,
                    peak_value=peak_value,
                    cumulative=cumulative,
                    fill_value=fill_value,
                )
            case ShapeType.block:
                return BlockTimeseries(
                    duration=duration,
                    peak_time=peak_time,
                    peak_value=peak_value,
                    cumulative=cumulative,
                    fill_value=fill_value,
                )
            case ShapeType.triangle:
                return TriangleTimeseries(
                    duration=duration,
                    peak_time=peak_time,
                    peak_value=peak_value,
                    cumulative=cumulative,
                    fill_value=fill_value,
                )
            case ShapeType.scs:
                if scs_file_name is None or scs_type is None:
                    from flood_adapt.dbs_classes.database import Database

                    scs_config = Database().site.sfincs.scs
                    if scs_config is None:
                        raise ValueError("SCS configuration not found in database.")
                    scs_file_name = scs_file_name or scs_config.file
                    scs_type = scs_type or scs_config.type
                return ScsTimeseries(
                    duration=duration,
                    peak_time=peak_time,
                    peak_value=peak_value,
                    cumulative=cumulative,
                    fill_value=fill_value,
                    scs_file_name=scs_file_name,
                    scs_type=scs_type,
                )
            case _:
                raise ValueError(f"Unknown shape type {shape_type}.")

    @staticmethod
    def load_file(
        file_path: Path | str | os.PathLike,
    ) -> SyntheticTimeseries:
        """Load object from file."""
        with open(file_path, mode="rb") as fp:
            toml = tomli.load(fp)
        return TimeseriesFactory.from_args(
            **toml,
        )

    @staticmethod
    def from_object(obj: SyntheticTimeseries) -> SyntheticTimeseries:
        return TimeseriesFactory.from_args(**obj.model_dump(exclude_none=True))



class FloodAdaptError(Exception):
    """Base class for all exceptions in the FloodAdapt module."""

    pass


class DatabaseError(FloodAdaptError):
    """Base class for exceptions raised in any database related files."""

    pass


class IsStandardObjectError(DatabaseError):
    """Raised when an operation is attempted on a standard object."""

    def __init__(
        self,
        name: str,
        object_type: str,
    ):
        msg = f"The {object_type} '{name}' is a standard object and cannot be modified or deleted."
        super().__init__(msg)


class AlreadyExistsError(DatabaseError):
    """Raised when an attempt is made to create an object that already exists."""

    def __init__(
        self,
        name: str,
        object_type: str,
    ):
        msg = f"The {object_type} '{name}' already exists."
        super().__init__(msg)


class DoesNotExistError(DatabaseError):
    """Raised when an attempt is made to access an object that does not exist."""

    def __init__(self, name: str, object_type: str):
        msg = f"The {object_type} '{name}' does not exist."
        super().__init__(msg)


class IsUsedInError(DatabaseError):
    """Raised when an attempt is made to delete or modify an object that is in use / referenced by another object."""

    def __init__(
        self, name: str, object_type: str, used_in_type: str, used_in: list[str]
    ):
        msg = f"The {object_type} '{name}' cannot be deleted/modified since it is already used in the {used_in_type}(s): {', '.join(used_in)}"
        super().__init__(msg)


class ConfigError(DatabaseError):
    """Raised when optional configuration, usually in the site, is missing or invalid."""

    def __init__(self, message: str):
        super().__init__(message)



import csv
from pathlib import Path

import pandas as pd


def read_csv(csvpath: Path) -> pd.DataFrame:
    """Read a timeseries file and return a pd.DataFrame.

    Parameters
    ----------
    csvpath : Path
        Path to the CSV file.

    Returns
    -------
    pd.DataFrame
        Dataframe with time as index and (a) data column(s).
    """
    num_columns = None
    has_header = None
    with open(csvpath, "r") as f:
        try:
            # read the first 1024 bytes to determine if there is a header
            has_header = csv.Sniffer().has_header(f.read(1024))
        except csv.Error:
            has_header = False
        f.seek(0)
        reader = csv.reader(f, delimiter=",")
        try:
            first_row = next(reader)
            num_columns = len(first_row) - 1  # subtract 1 for the index column
        except StopIteration:
            raise ValueError(f"The CSV file is empty: {csvpath}.")

    if has_header is None:
        raise ValueError(
            f"Could not determine if the CSV file has a header: {csvpath}."
        )
    if num_columns is None:
        raise ValueError(
            f"Could not determine the number of columns in the CSV file: {csvpath}."
        )
    elif num_columns < 1:
        raise ValueError(f"CSV file must have at least one data column: {csvpath}.")
    columns = [f"data_{i}" for i in range(num_columns)]
    dtype = {name: float for name in columns}

    df = pd.read_csv(
        csvpath,
        index_col=0,
        names=columns,
        header=0 if has_header else None,
        parse_dates=True,
        infer_datetime_format=True,
        dtype=dtype,
    )

    # Any index that cannot be converted to datetime will be NaT
    df.index = pd.to_datetime(df.index, errors="coerce")
    df.index.names = ["time"]
    if len(df.index) > 2:
        df.index.freq = pd.infer_freq(df.index)

    # Drop rows where the index is NaT
    df = df[~df.index.isna()]

    return df



from pathlib import Path
from typing import Optional

from pydantic import BaseModel, model_validator
from tomli import load as load_toml

from flood_adapt.config.hazard import (
    Cstype,
    CycloneTrackDatabaseModel,
    DemModel,
    FloodFrequencyModel,
    FloodModel,
    ObsPointModel,
    RiverModel,
    SCSModel,
    SlrScenariosModel,
    WaterlevelReferenceModel,
)
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.tide_gauge import TideGauge


class SfincsConfigModel(BaseModel):
    """The expected variables and data types of attributes of the SfincsConfig class.

    Attributes
    ----------
    csname : str
        The name of the CS model.
    cstype : Cstype
        Cstype of the CS model. must be either "projected" or "spherical".
    version : Optional[str], default = None
        The version of the CS model. If None, the version is not specified.
    offshore_model : Optional[FloodModel], default = None
        The offshore model. If None, the offshore model is not specified.
    overland_model : FloodModel
        The overland model. This is the main model used for the simulation.
    floodmap_units : us.UnitTypesLength
        The units used for the output floodmap. Sfincs always produces in metric units, this is used to convert the floodmap to the correct units.
    save_simulation : Optional[bool], default = False
        Whether to keep or delete the simulation files after the simulation is finished and all output files are created.
        If True, the simulation files are kept. If False, the simulation files are deleted.
    """

    csname: str
    cstype: Cstype
    version: Optional[str] = None
    offshore_model: Optional[FloodModel] = None
    overland_model: FloodModel
    floodmap_units: us.UnitTypesLength
    save_simulation: Optional[bool] = False


class SfincsModel(BaseModel):
    """The expected variables and data types of attributes of the Sfincs class.

    Attributes
    ----------
    config : SfincsConfigModel
        The configuration of the Sfincs model.
    water_level : WaterlevelReferenceModel
        The collection of all datums and the main reference datum.
    dem : DemModel
        The digital elevation model.
    flood_frequency : FloodFrequencyModel, default = FloodFrequencyModel()
        The flood frequency model.
    slr : SlrScenariosModel
        Specification of the sea level rise scenarios.
    cyclone_track_database : CycloneTrackDatabaseModel, optional, default = None
        The cyclone track database model.
    scs : SCSModel, optional, default = None
        The SCS model.
    tide_gauge : TideGauge, optional, default = None
        The tide gauge model.
    river : list[RiverModel], optional, default = None
        The river model.
    obs_point : list[ObsPointModel], optional, default = None
        The observation point model.
    """

    config: SfincsConfigModel
    water_level: WaterlevelReferenceModel
    cyclone_track_database: Optional[CycloneTrackDatabaseModel] = None
    slr_scenarios: Optional[SlrScenariosModel] = None
    scs: Optional[SCSModel] = None  # optional for the US to use SCS rainfall curves
    dem: DemModel

    flood_frequency: FloodFrequencyModel = FloodFrequencyModel(
        flooding_threshold=us.UnitfulLength(value=0.0, units=us.UnitTypesLength.meters)
    )  # TODO we dont actually use this anywhere?

    tide_gauge: Optional[TideGauge] = None
    river: Optional[list[RiverModel]] = None
    obs_point: Optional[list[ObsPointModel]] = None

    @staticmethod
    def read_toml(path: Path) -> "SfincsModel":
        with open(path, mode="rb") as fp:
            toml_contents = load_toml(fp)

        return SfincsModel(**toml_contents)

    @model_validator(mode="after")
    def ensure_references_exist(self):
        datum_names = [d.name for d in self.water_level.datums]

        if self.config.overland_model.reference not in datum_names:
            raise ValueError(
                f"Could not find reference `{self.config.overland_model.reference}` in available datums: {datum_names}."
            )

        if self.config.offshore_model is not None:
            if self.config.offshore_model.reference not in datum_names:
                raise ValueError(
                    f"Could not find reference `{self.config.offshore_model.reference}` in available datums: {datum_names}."
                )

        if self.tide_gauge is not None:
            if self.tide_gauge.reference not in datum_names:
                raise ValueError(
                    f"Could not find reference `{self.tide_gauge.reference}` in available datums: {datum_names}."
                )

        return self



import os
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Any, List, Type

import tomli
from pydantic import BaseModel, field_serializer

from flood_adapt.config.hazard import RiverModel


### ENUMS ###
class ForcingType(str, Enum):
    """Enum class for the different types of forcing parameters.

    Attributes
    ----------
    RAINFALL : The type of forcing parameter for rainfall.
    WIND : The type of forcing parameter for wind.
    DISCHARGE : The type of forcing parameter for discharge.
    WATERLEVEL : The type of forcing parameter for water level.
    """

    WIND = "WIND"
    RAINFALL = "RAINFALL"
    DISCHARGE = "DISCHARGE"
    WATERLEVEL = "WATERLEVEL"


class ForcingSource(str, Enum):
    """Enum class for the different sources of forcing parameters."""

    MODEL = "MODEL"  # 'our' hindcast/ sfincs offshore model
    TRACK = "TRACK"  # 'our' hindcast/ sfincs offshore model + (shifted) hurricane
    CSV = "CSV"  # user provided csv file
    NETCDF = "NETCDF"  # user provided netcdf file

    SYNTHETIC = "SYNTHETIC"  # synthetic data
    CONSTANT = "CONSTANT"  # synthetic data

    GAUGED = "GAUGED"  # data downloaded from a gauge
    METEO = "METEO"  # external hindcast data

    NONE = "NONE"  # no forcing data


class IForcing(BaseModel, ABC):
    """BaseModel describing the expected variables and data types for forcing parameters of hazard model."""

    class Config:
        arbitrary_types_allowed = True

    type: ForcingType
    source: ForcingSource

    @classmethod
    def load_file(cls, path: Path):
        with open(path, mode="rb") as fp:
            toml_data = tomli.load(fp)
        return cls.load_dict(toml_data)

    @classmethod
    def load_dict(cls, attrs):
        return cls.model_validate(attrs)

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """Override the default model_dump to include class variables `type` and `source`."""
        data = super().model_dump(**kwargs)
        data.update({"type": self.type, "source": self.source})
        return data

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        """Save additional data of the forcing."""
        return

    @field_serializer("path", check_fields=False)
    @classmethod
    def serialize_path(cls, value: Path) -> str:
        """Serialize filepath-like fields by saving only the filename. It is assumed that the file will be saved in the same directory."""
        return value.name


class IDischarge(IForcing):
    type: ForcingType = ForcingType.DISCHARGE
    river: RiverModel


class IRainfall(IForcing):
    type: ForcingType = ForcingType.RAINFALL


class IWind(IForcing):
    type: ForcingType = ForcingType.WIND


class IWaterlevel(IForcing):
    type: ForcingType = ForcingType.WATERLEVEL


class IForcingFactory:
    @classmethod
    @abstractmethod
    def load_file(cls, toml_file: Path) -> IForcing:
        """Create a forcing object from a TOML file."""
        ...

    @classmethod
    @abstractmethod
    def load_dict(cls, attrs: dict[str, Any] | IForcing) -> IForcing:
        """Create a forcing object from a dictionary of attributes."""
        ...

    @classmethod
    @abstractmethod
    def read_forcing(
        cls,
        filepath: Path,
    ) -> tuple[Type[IForcing], ForcingType, ForcingSource]:
        """Extract forcing class, type and source from a TOML file."""
        ...

    @classmethod
    @abstractmethod
    def get_forcing_class(
        cls, type: ForcingType, source: ForcingSource
    ) -> Type[IForcing]:
        """Get the forcing class corresponding to the type and source."""
        ...

    @classmethod
    @abstractmethod
    def list_forcing_types(cls) -> List[ForcingType]:
        """List all available forcing types."""
        ...

    @classmethod
    @abstractmethod
    def list_forcing_classes(cls) -> List[Type[IForcing]]:
        """List all available forcing classes."""
        ...

    @classmethod
    @abstractmethod
    def list_forcing_types_and_sources(cls) -> List[tuple[ForcingType, ForcingSource]]:
        """List all available combinations of forcing types and sources."""
        ...



from abc import ABC


class DatabaseUser(ABC):
    """Abstract class for FloodAdapt classes that need to use / interact with the Singleton FloodAdapt database through the lazy-loading self.database property."""

    _database_instance = None

    @property
    def database(self):
        """Return the database for the object."""
        if self._database_instance is None:
            from flood_adapt.dbs_classes.database import Database

            self._database_instance = Database()
        return self._database_instance



from flood_adapt.config.config import Settings
from flood_adapt.config.fiat import (
    FiatConfigModel,
    FiatModel,
)
from flood_adapt.config.gui import (
    AggregationDmgLayer,
    BenefitsLayer,
    FloodMapLayer,
    FootprintsDmgLayer,
    GuiModel,
    GuiUnitModel,
    Layer,
    OutputLayers,
    PlottingModel,
    VisualizationLayer,
    VisualizationLayers,
)
from flood_adapt.config.hazard import (
    AsciiStr,
    Cstype,
    CycloneTrackDatabaseModel,
    DatumModel,
    DemModel,
    FloodFrequencyModel,
    FloodModel,
    ObsPointModel,
    RiverModel,
    SCSModel,
    Scstype,
    SlrScenariosModel,
    WaterlevelReferenceModel,
)
from flood_adapt.config.impacts import (
    AggregationModel,
    BenefitsModel,
    EquityModel,
    FloodmapType,
    NoFootprintsModel,
    RiskModel,
)
from flood_adapt.config.sfincs import (
    SfincsConfigModel,
    SfincsModel,
)
from flood_adapt.config.site import Site, SiteBuilder, StandardObjectModel

__all__ = [
    # GUI
    "GuiModel",
    "GuiUnitModel",
    "PlottingModel",
    "Layer",
    "OutputLayers",
    "BenefitsLayer",
    "FloodMapLayer",
    "FootprintsDmgLayer",
    "VisualizationLayer",
    "VisualizationLayers",
    "AggregationDmgLayer",
    # FIAT
    "FiatModel",
    "FiatConfigModel",
    "FloodmapType",
    "NoFootprintsModel",
    "RiskModel",
    "EquityModel",
    "BenefitsModel",
    "AggregationModel",
    # Config
    "Settings",
    # Sfincs
    "SfincsModel",
    "SfincsConfigModel",
    "SCSModel",
    "Scstype",
    "Cstype",
    "DemModel",
    "DatumModel",
    "FloodModel",
    "RiverModel",
    "ObsPointModel",
    "SlrScenariosModel",
    "FloodFrequencyModel",
    "WaterlevelReferenceModel",
    "CycloneTrackDatabaseModel",
    "AsciiStr",
    # Site
    "Site",
    "SiteBuilder",
    "StandardObjectModel",
]



from pathlib import Path
from typing import Optional

from pydantic import BaseModel
from tomli import load as load_toml

from flood_adapt.config.impacts import (
    AggregationModel,
    BenefitsModel,
    BFEModel,
    FloodmapType,
    NoFootprintsModel,
    RiskModel,
    SVIModel,
)


class FiatConfigModel(BaseModel):
    """Configuration settings for the FIAT model.

    Attributes
    ----------
    exposure_crs : str
        The coordinate reference system of the exposure data.
    bfe : Optional[BFEModel], default=None
        The base flood elevation model.
    aggregation : list[AggregationModel]
        Configuration for the aggregation model.
    floodmap_type : FloodmapType
        The type of flood map to be used.
    non_building_names : Optional[list[str]], default=None
        List of non-building names to be used in the model.
    damage_unit : str, default="$"
        The unit of damage used in the model.
    building_footprints : Optional[str], default=None
        Path to the building footprints data.
    roads_file_name : Optional[str], default=None
        Path to the roads data.
    new_development_file_name : Optional[str], default="new_development_area.gpkg"
        Path to the new development area data.
    save_simulation : Optional[bool], default=False
        Whether to keep or delete the simulation files after the simulation is finished and all output files are created.
        If True, the simulation files are kept. If False, the simulation files are deleted.
    svi : Optional[SVIModel], default=None
        The social vulnerability index model.
    infographics : Optional[bool], default=False
        Whether to create infographics or not.
    no_footprints : Optional[NoFootprintsModel], default=NoFootprintsModel()
        Configuration for objects with no footprints.
    """

    exposure_crs: str
    bfe: Optional[BFEModel] = None
    aggregation: list[AggregationModel]
    floodmap_type: FloodmapType
    non_building_names: Optional[list[str]]
    damage_unit: str = "$"
    building_footprints: Optional[str] = None
    roads_file_name: Optional[str] = None
    new_development_file_name: Optional[str] = "new_development_area.gpkg"
    save_simulation: Optional[bool] = False
    svi: Optional[SVIModel] = None
    infographics: Optional[bool] = False
    no_footprints: Optional[NoFootprintsModel] = NoFootprintsModel()

    @staticmethod
    def read_toml(path: Path) -> "FiatConfigModel":
        with open(path, mode="rb") as fp:
            toml_contents = load_toml(fp)

        return FiatConfigModel(**toml_contents)


class FiatModel(BaseModel):
    """The expected variables and data types of attributes of the Fiat class.

    Attributes
    ----------
    risk : Optional[RiskModel]
        Configuration of probabilistic risk runs. default=None
    config : FiatConfigModel
        Configuration for the FIAT model.
    benefits : Optional[BenefitsModel]
        Configuration for running benefit calculations. default=None
    """

    config: FiatConfigModel

    benefits: Optional[BenefitsModel] = None
    risk: RiskModel = RiskModel()

    @staticmethod
    def read_toml(path: Path) -> "FiatModel":
        with open(path, mode="rb") as fp:
            toml_contents = load_toml(fp)

        return FiatModel(**toml_contents)
